//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  mzscript.cpp
//  This is a part of the mzscript source code.
//
//  Copyright (C) 2000-2003 the mzscript developers.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

#include "bangs.hpp"
#include "scripts.hpp"
#include "utility.hpp"
#include "mzscript.hpp"

using std::string;
using std::invalid_argument;
using std::runtime_error;


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// local helpers
//
static bool ParseScript(LPCSTR pszName, FILE* fStart);


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Globals
//
HWND g_hMain = NULL;
HWND g_hLitestep = NULL;

const UINT g_lsMessages[] = { LM_GETREVID, LM_REFRESH, 0 };


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// initModuleEx
//
int initModuleEx(HWND hLiteStep, HINSTANCE hInst, LPCSTR /* pszPath */)
{
    g_hLitestep = hLiteStep;

    try
    {
        //
        // window (class) creation/registration
        //
        WNDCLASS wc = { 0 };
        wc.lpfnWndProc = WndProc;
        wc.hInstance = hInst;
        wc.lpszClassName = g_szAppName;
        
        if (!RegisterClass(&wc))
        {
            throw win32_error();
        }

        g_hMain = CreateMessageWindow(hInst, g_szAppName, NULL);

        SetWindowLong(g_hMain, GWL_USERDATA, magicDWord);
        SendMessage(g_hLitestep, LM_REGISTERMESSAGE,
            (WPARAM)g_hMain, (LPARAM)g_lsMessages);

        //
        // register bangs, read step.rc settings
        //
        RegisterBangs();
        
        LoadSetup(false);
    }
    catch (std::bad_alloc&)
    {
        ErrorMessage("Bad allocation during initialization.\n"
            "Out of memory?\n\nShutting down...");
        
        ClearSetup(); // could also use quitModule, but that might be a bit more
                      // dangerous... then again you can hardly do anything
                      // "right" in OOM situations.
        return 2;
    }
    catch (std::exception& e)
    {
        ErrorMessage("Initialization failed.\n\nDescription: %s", e.what());

        return 1;
    }

    return 0;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// quitModule
//
void quitModule(HINSTANCE hInst)
{
    UnregisterBangs();
    ClearSetup();
    
    if (g_hMain)
    {
        SendMessage(g_hLitestep, LM_UNREGISTERMESSAGE,
            (WPARAM)g_hMain, (LPARAM)g_lsMessages);
        
        DestroyWindow(g_hMain);

        g_hMain = NULL;
    }

    UnregisterClass(g_szAppName, hInst);

    g_hLitestep = NULL;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// LoadSetup
//
// throws invalid_argument, runtime_error
//
void LoadSetup(bool bIsRefresh)
{
    LoadScriptFile(NULL, bIsRefresh);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ClearSetup
//
// must not throw
//
void ClearSetup()
{
    ScriptsClear();
    g_Variables->Clear();
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// LoadScriptFile
//
// throws invalid_argument, runtime_error
//
void LoadScriptFile(LPCSTR pszFile, bool bIsRefresh)
{
    FILE* f = LCOpen(pszFile);

    if (f)
    {
        char szBuffer[MAX_LINE_LENGTH] = { 0 };

        while (LCReadNextConfig(f, "*Script", szBuffer, MAX_LINE_LENGTH))
        {
            char szType[MAX_LINE_LENGTH] = { 0 };
            char szRest[MAX_LINE_LENGTH] = { 0 };
            char* tokens[] = { NULL, szType };

            int nCount = LCTokenize(szBuffer, tokens, COUNTOF(tokens), szRest);

            if (nCount >= 2 && szRest[0])
            {
                if ((StrIEqual(szType, "start") && !bIsRefresh) ||
                    (StrIEqual(szType, "refresh") && bIsRefresh))
                {
                    ExecCommand(szRest);
                }
                else if (StrIEqual(szType, "bang"))
                {
                    if (!ParseScript(szRest, f))
                    {
                        break;
                    }
                }
                else
                {
                    int nResponse = AskUser(MB_YESNO | MB_ICONWARNING,
                        "Invalid or misplaced token \"%s\":\n"
                        "*Script %s %s\n\nContinue parsing?",
                        szType, szType, szRest);

                    if (nResponse == IDNO)
                    {
                        break;
                    }
                }
            }
        }
    }
    else
    {
        FormattedException<runtime_error>("Could not open \"%s\"",
            pszFile ? pszFile : "step.rc");
    }

    LCClose(f); // FIXME: not exception safe
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ParseScript
//
// Process all lines between *Script bang and *Script ~bang
// throws invalid_argument
//
static bool ParseScript(LPCSTR pszName, FILE* fStart)
{
    ASSERT(pszName); ASSERT(pszName[0]);
    
    ScriptPtr pScript(new Script(pszName));
    
    bool bCompleted = false;
    char szBuffer[MAX_LINE_LENGTH] = { 0 };
    
    while (LCReadNextConfig(fStart, "*Script", szBuffer, MAX_LINE_LENGTH))
    {
        try
        {
            char szType[MAX_LINE_LENGTH] = { 0 };
            char szRest[MAX_LINE_LENGTH] = { 0 };
            
            // NULL skips the "*Script" token
            char* ppszTokens[] = { NULL, szType };
            
            if (LCTokenize(szBuffer, ppszTokens, COUNTOF(ppszTokens), szRest) >= 2)
            {
                if (StrIEqual(szType, "~bang"))
                {
                    bCompleted = true;
                    break;
                }
                
                pScript->AddLine(szType, szRest);
            }
            else
            {
                throw invalid_token(szType);
            }
        }
        catch (invalid_token& it)
        {
            int nResponse = AskUser(MB_YESNO | MB_ICONWARNING,
                "Invalid or misplaced token \"%s\" in script \"%s\"\n\n"
                "Continue parsing?",
                it.what(), pszName);
            
            if (nResponse == IDNO)
            {
                return false;
            }
        }
        catch (invalid_argument& ia)
        {
            int nResponse = AskUser(MB_YESNO | MB_ICONWARNING,
                "Invalid argument in script \"%s\"\n%s\n\n"
                "Continue parsing?",
                ia.what(), pszName);
            
            if (nResponse == IDNO)
            {
                return false;
            }
        }
    }
    
    if (!bCompleted)
    {
        FormattedException<invalid_argument>(
            "Script %s has no closing ~bang line", pszName);
    }

    ScriptAdd(pScript);

    return true;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// EvalExpression
//
static bool EvalExpression(const string& sExpression)
{
    char szFirst[MAX_LINE_LENGTH] = { 0 };
    char szComp[MAX_LINE_LENGTH] = { 0 };
    char szSecond[MAX_LINE_LENGTH] = { 0 };

    char* ppszTokens[] = { szFirst, szComp, szSecond };
    LCTokenize(sExpression.c_str(), ppszTokens, COUNTOF(ppszTokens), NULL);

    boost::function2<bool, int, int> fn = std::not_equal_to<int>();

    if (StrIEqual(szComp, "="))
    {
        fn = std::equal_to<int>();
    }
    else if (StrIEqual(szComp, "<="))
    {
        fn = std::less_equal<int>();
    }
    else if (StrIEqual(szComp, ">="))
    {
        fn = std::greater_equal<int>();
    }
    else if (StrIEqual(szComp, "<"))
    {
        fn = std::less<int>();
    }
    else if (StrIEqual(szComp, ">"))
    {
        fn = std::greater<int>();
    }
    else if (StrIEqual(szComp, "<>"))
    {
        fn = std::not_equal_to<int>();
    }

    if (isnum(szFirst) && isnum(szSecond))
    {
        if (fn(atoi(szFirst), atoi(szSecond)))
        {
            return true;
        }
    }
    else
    {
        if (fn(StrIComp(szFirst, szSecond), 0))
        {
            return true;
        }
    }

    return false;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// EvalIf
//
bool EvalIf(std::string& sStatement)
{
    using std::invalid_argument;
    
    bool bReturn = false;
    char szCondition[MAX_LINE_LENGTH] = { 0 };
    char szTrue[MAX_LINE_LENGTH] = { 0 };
    char szFalse[MAX_LINE_LENGTH] = { 0 };
    
    char* ppszTokens[] = { szCondition, szTrue, szFalse };
    
    if (CommandTokenize(sStatement.c_str(), ppszTokens, COUNTOF(ppszTokens), NULL) >= 2)
    {
        if (!szCondition[0])
        {
            throw invalid_argument("No condition in if-statement given");
        }
        
        if (!szTrue[0] && !szFalse[0])
        {
            throw invalid_argument("if-statement has neither actions for "
                "\"true\" nor for \"false\"");
        }
        
        bReturn = EvalExpression(szCondition);
        sStatement.assign(bReturn ? szTrue : szFalse);
    }
    else
    {
        throw invalid_argument("Not enough tokens in if-statement: " +
            sStatement);
    }
    
    return bReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ExecCommand
//
void ExecCommand(const std::string& sCommand)
{
    ASSERT(!sCommand.empty());
    LSExecute(NULL, sCommand.c_str(), 0);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// WndProc
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
    switch (uMessage)
    {
        case LM_GETREVID:
        {
            if (wParam == 0 || wParam == 1)
            {
                if (SUCCEEDED(StringCchPrintf((char*)lParam, 64,
                    "%s %s", g_szAppName, g_szAppVersion)))
                {
                    return lstrlen((char*)lParam);
                }
            }
            
            // something above failed
            lParam = NULL;
        }
        break;

        case LM_REFRESH:
        {
            ClearSetup();
            LoadSetup(true);

            return 0;
        }
        break;

        default:
        break;
    }

    return DefWindowProc(hWnd, uMessage, wParam, lParam);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// DllMain
//
// DLL entry point
//
BOOL APIENTRY DllMain(HINSTANCE hInst, DWORD dwReason, LPVOID /* pvReserved */)
{
    if (dwReason == DLL_PROCESS_ATTACH)
    {
        DisableThreadLibraryCalls(hInst);
    }
    
    return TRUE;
}
