//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  variables.cpp
//  This is a part of the mzscript source code.
//
//  Copyright (C) 2000-2003 the mzscript developers.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

#include "utility.hpp"
#include "variables.hpp"

using std::string;


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Globals/Static members
//
//    
Variables::Ptr g_Variables(new Variables(Variables::Ptr()));


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variable::Get
//
string Variable::Get(const string& sIndex) const
{
    string sReturn;

    if (m_pFunc)
    {
        m_sValue = m_pFunc();
    }
    
    if (isnum(sIndex))
    {
        int nIndex = atoi(sIndex.c_str());

        if (nIndex > 0)
        {
            //
            // array handling
            //
            size_t stStart, stEnd;
            if (_GetElementPos(nIndex, stStart, stEnd) == nIndex)
            {
                sReturn.assign(m_sValue.substr(stStart, stEnd-stStart));
            }
        }
    }
    else if (StrIEqual(sIndex, "count"))
    {
        size_t stDummy1, stDummy2;
        sReturn = string_cast(_GetElementPos(INT_MAX, stDummy1, stDummy2));
    }
    else if (StrIEqual(sIndex, "file"))
    {
        sReturn.assign(m_sFile);
    }
    else if (StrIEqual(sIndex, "sep"))
    {
        sReturn = m_cSeparator;
    }
    else
    {
        sReturn.assign(m_sValue);
    }
    
    return sReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variable::Set
//
void Variable::Set(const string& sValue, const string& sIndex)
{
    if (isnum(sIndex))
    {
        int nIndex = atoi(sIndex.c_str());
        
        if (nIndex > 0)
        {
            //
            // array handling
            //
            size_t stStart, stEnd;
            int nElements = _GetElementPos(nIndex, stStart, stEnd);
            
            if (nElements != nIndex)
            {
                m_sValue.append(nIndex - nElements, m_cSeparator);
                stStart = stEnd = m_sValue.length();
            }
            
            m_sValue.replace(stStart, stEnd-stStart, sValue);
        }
    }
    else if (StrIEqual(sIndex, "sep"))
    {
        m_cSeparator = sValue[0];
    }
    else
    {
        //
        // change entire value
        //
        m_sValue = sValue;
    }
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variable::Remove
//
// Returns true if the variable can be deleted
//
bool Variable::Remove(const string& sIndex)
{
    if (isnum(sIndex))
    {
        int nIndex = atoi(sIndex.c_str());

        if (nIndex > 0)
        {
            size_t stStart, stEnd;
            
            if (_GetElementPos(nIndex, stStart, stEnd) == nIndex)
            {
                if (nIndex > 1)
                {
                    // need to delete the leading separator
                    --stStart;
                }
                else if (stEnd != string::npos)
                {
                    // need to delete the trailing separator
                    ++stEnd;
                }
                
                m_sValue.erase(stStart, stEnd-stStart);
            }
        }
        
        return false;
    }
    else if (StrIEqual(sIndex, "sep"))
    {
        m_cSeparator = '\0';
    }
    else if (!sIndex.empty())
    {
        return false;
    }

    return true;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variable::_GetElementPos
//
// Get the position of element nIndex in the array
//
int Variable::_GetElementPos(int nIndex, size_t& stStart, size_t& stEnd) const
{
    ASSERT(nIndex > 0);

    int nCounter = nIndex;

    if (m_cSeparator == '\0')
    {
        //
        // iterate through the string if separator is empty
        //
        if (nIndex > m_sValue.length())
        {
            nIndex = m_sValue.length();
        }

        stStart = nIndex-1;
        stEnd = nIndex;

        nCounter = 0;
    }
    else
    {
        bool bInQuote = false;
        
        string::const_iterator iter, itStart, itEnd;

        iter = itEnd = m_sValue.begin();
        itStart = m_sValue.end();
        --itEnd;

        do
        {
            if (iter == m_sValue.end() || (*iter == m_cSeparator && !bInQuote))
            {
                itStart = itEnd;
                itEnd = iter;
                --nCounter;
            }
            else if (*iter == '\"')
            {
                bInQuote = !bInQuote;
            }
        }
        while (iter++ != m_sValue.end() && nCounter);

        stStart = itStart - m_sValue.begin() + 1;
        stEnd = itEnd - m_sValue.begin();
    }

    return (nIndex - nCounter);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::Get
//
bool Variables::Get(string sVariable, string* psValue) const
{
    assert(!sVariable.empty());
    string sIndex = _InterpretVar(sVariable);

    return _GetValueWorker(sVariable, psValue, sIndex);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::GetFile
//
string Variables::GetFile(const string& sVariable) const
{
    return "unknown :p";
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::_GetValueWorker
//
bool Variables::_GetValueWorker(const string& sVariable, string* psValue,
                                const string& sIndex) const
{
    VarMap::const_iterator iter = m_VarMap.find(sVariable);
    
    if (iter != m_VarMap.end())
    {
        ASSERT(iter->second);
        
        if (psValue)
        {
            *psValue = iter->second->Get(sIndex);
        }
        
        return true;
    }
    
    if (m_pParent)
    {
        return m_pParent->_GetValueWorker(sVariable, psValue, sIndex);
    }
    
    return false;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::Set
//
// Change a variable's value or create a new variable
//
Variables::VarPtr Variables::Set(string sVariable, const string& sValue,
                                 const string& sFile)
{
    assert(!sVariable.empty());
    
    string sIndex = _InterpretVar(sVariable);
    VarMap::iterator iter = m_VarMap.find(sVariable);

    if (iter != m_VarMap.end())
    {
        ASSERT(iter->second);

        // variable already exists, change its value
        iter->second->Set(sValue, sIndex);
        
        return iter->second;
    }
    else
    {
        // create a new variable
        VarPtr pVar = VarPtr(new Variable(sValue, sFile));
        
        if (pVar)
        {
            m_VarMap.insert(VarMap::value_type(sVariable, pVar));
            
            return pVar;
        }
    }

    ErrorMessage("Couldn't set variable \"%s\" to value \"%s\".\n"
        "Out of memory?", sVariable.c_str(), sValue.c_str());
    
    return VarPtr();
}

bool Variables::Set(const string& sVariable, VarFunc pFunc)
{
    // TODO: implement var saving if var was updated
    ASSERT(!sVariable.empty());
    
    m_VarMap.erase(sVariable);
    
    VarPtr pVar = VarPtr(new Variable(pFunc));
    
    if (pVar)
    {
        m_VarMap.insert(VarMap::value_type(sVariable, pVar));
        return true;
    }
    
    ErrorMessage("Couldn't set function variable \"%s\".\nOut of memory?",
        sVariable.c_str());
    
    return false;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Variables::Remove
//
bool Variables::Remove(string sVariable)
{
    ASSERT(!sVariable.empty());
    
    string sIndex = _InterpretVar(sVariable);

    VarMap::iterator iter = m_VarMap.find(sVariable);
    
    if (iter != m_VarMap.end())
    {
        if (iter->second->Remove(sIndex))
        {
            m_VarMap.erase(iter);
        }

        return true;
    }
    
    return false;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// _InterpretVar
//
string Variables::_InterpretVar(string& sVariable) const
{
    string sReturn;
    
    size_t stIndex = sVariable.find(':');
    
    if (stIndex != string::npos && sVariable.length() > stIndex+1)
    {
        sReturn.assign(&sVariable.c_str()[stIndex+1]);
        sVariable.erase(stIndex, sVariable.length()-stIndex);
    }
    
    return sReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Expand
//
bool Variables::Expand(string& sString) const
{
    bool bReturn = false;
    
    if (!sString.empty())
    {
        // let the worker do the real "work" =)
        bReturn = _ExpandVarsWorker(sString);
    }
    
    return bReturn;
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ExpandVarsWorker
//
// Worker function for Expand (see above).
// This is probably rather slow since it uses std::basic_string functions
// instead of c-string manipulation, but it's a lot easier to maintain this way
// and most likely efficient enough for now.
//
bool Variables::_ExpandVarsWorker(string& sString) const
{
    const char cEscape = '%';
    
    size_t stPrefix = sString.rfind(cEscape);
    
    // this also catches empty strings
    if (stPrefix == string::npos)
    {
        // nothing to expand
        return false;
    }

    if (stPrefix >= sString.length() - 1)
    {
        ErrorMessage("Escape code character found at end of line\n"
            "Expression: %s", sString.c_str());

        return false;
    }

    //
    // actual expansion starts here
    //
    switch(sString.at(stPrefix+1))
    {
        //
        // variable/array
        //
        case '{':
        {
            size_t stPostfix = sString.find('}', stPrefix+2);
            
            if (stPostfix != string::npos)
            {
                // Now we know the position of the %{ and the closing }.
                // Construct a new string which holds the variable's name.
                string sVariable(sString, stPrefix+2, stPostfix-stPrefix-2);
                string sValue;
                
                if (Get(sVariable, &sValue))
                {
                    sString.replace(stPrefix, stPostfix-stPrefix+1, sValue);
                }
                else
                {
                    // replace with empty string if variable doesn't exist
                    sString.erase(stPrefix, stPostfix-stPrefix+1);
                }

            }
            else
            {
                ErrorMessage("Closing } not found.\nExpression: %s",
                    sString.c_str());

                return false;
            }
        }
        break;
        
        //
        // escape codes
        //
        case '#':
        {
            sString.replace(stPrefix, 2, "$");
        }
        break;
            
        case '=':
        {
            sString.replace(stPrefix, 2, "\"");
        }
        break;
            
        default:
        {
            ErrorMessage("Unknown escape code '%c'.\nExpression: %s",
                sString.at(stPrefix+1), sString.c_str());
            
            sString.erase(stPrefix, 2);
        }
        break;
    }
    
    // repeat as long as there could be more to expand
    _ExpandVarsWorker(sString);
    
    return true;
}
